# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"
__version__ = "0.0.1"


class Tables:
    EXPERIMENTS = 'experiments'
    SESSIONS = 'sessions'
    LOGS = 'perf_logs'
    MODELS = 'models'
    PREDICTIONS = 'predictions'


class ConfigRef:
    ARCHITECTURE = 'cf_architecture'
    DATASET_PARAMS = 'cf_dataset_params'
    INPUT_PARAMS = 'cf_input_params'
    MODEL_PARAMS = 'cf_model_params'
    TRAIN_PARAMS = 'cf_train_params'


class Providers:
    MONGODB = 'MongoDB'
    MYSQL = 'MySql'


class LogAttr:
    SESSION_ID = 'session'
    EPOCH = 'epoch'


class SessionAttr:
    EXPERIMENT_ID = 'experiment'
    CREATED_TIME = 'created'
    LAST_MODIFIED = 'lastModified'
    CURRENT_EPOCH = 'curr_epoch'
    STATUS = 'status'


class HDF5Attr:
    SESSION_ID = 'session'
    EPOCH = 'epoch'
    FILE_LOCATION = 'location'


class ExperimentAttr:
    CONFIG = 'config'
    SAVED_MODEL_LOC = 'file_location'
    NAME = 'name'
    DESC = 'description'
    REF_ARCHITECTURE = 'architecture'
    REF_DATASET_PARAMS = 'dataset_params'
    REF_INPUT_PARAMS = 'input_params'
    REF_MODEL_PARAMS = 'model_params'
    REF_TRAIN_PARAMS = 'train_params'


class RefAttr:
    NAME = 'name'
    CONFIG = 'config'
    DESC = 'description'


class SessionStatus:
    CREATED = 'created'
    TRAINING = 'training'
    FINISHED = 'finished'
    FAILED = 'failed'



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"
__version__ = "0.0.1"


import pytest
import numpy as np
from deoxys.data import Preprocessors, BasePreprocessor, \
    SingleChannelPreprocessor, WindowingPreprocessor, \
    KerasImagePreprocessorX, KerasImagePreprocessorY, \
    preprocessor_from_config
from deoxys.customize import register_preprocessor, \
    unregister_preprocessor, custom_preprocessor
from deoxys.utils import Singleton


@pytest.fixture(autouse=True)
def clear_singleton():
    Singleton._instances = {}  # clear singleton


@pytest.fixture
def preprocessor_class():
    class TestPreprocessor(BasePreprocessor):
        pass

    yield TestPreprocessor


def test_is_singleton():
    preprocessors_instance = Preprocessors()
    another_instance = Preprocessors()

    assert preprocessors_instance is another_instance


def test_register_random_obj():
    with pytest.raises(ValueError):
        register_preprocessor("TestPreprocessor", object)


def test_register_preprocessor_success(preprocessor_class):
    register_preprocessor("TestPreprocessor", preprocessor_class)

    assert Preprocessors(
    )._preprocessors["TestPreprocessor"] is preprocessor_class


def test_register_duplicate_preprocessor(preprocessor_class):
    register_preprocessor("TestPreprocessor", preprocessor_class)

    with pytest.raises(KeyError):
        register_preprocessor("TestPreprocessor", preprocessor_class)


def test_unregister_preprocessor(preprocessor_class):
    register_preprocessor("TestPreprocessor", preprocessor_class)

    assert Preprocessors(
    )._preprocessors["TestPreprocessor"] is preprocessor_class
    unregister_preprocessor("TestPreprocessor")

    assert "TestPreprocessor" not in Preprocessors()._preprocessors


def test_decorator():
    @custom_preprocessor
    class TestPreprocessor2(BasePreprocessor):
        pass

    assert Preprocessors(
    ).preprocessors["TestPreprocessor2"] is TestPreprocessor2


def test_preprocessor_from_config():
    pp = preprocessor_from_config({
        'class_name': 'SingleChannelPreprocessor'
    })

    assert callable(pp.transform)


def test_single_channel_preprocessor():
    scp = SingleChannelPreprocessor()

    x, y = np.arange(100).reshape(4, 5, 5), np.arange(100).reshape(4, 5, 5)
    expected_x = np.arange(100).reshape(4, 5, 5, 1)
    expected_y = np.arange(100).reshape(4, 5, 5, 1)

    output_x, output_y = scp.transform(x, y)
    assert x.shape == (4, 5, 5), 'X remains unchanged'
    assert y.shape == (4, 5, 5), 'Y remains unchanged'
    assert expected_x.shape == output_x.shape
    assert expected_y.shape == output_y.shape
    assert np.all(expected_x == output_x)
    assert np.all(expected_y == output_y)


def test_windowing_preprocessor():
    wp = WindowingPreprocessor(50, 80, 0)

    y = np.arange(200).reshape(4, 5, 5, 2)
    x_first_channel = np.arange(100).reshape(4, 5, 5)
    x_second_channel = np.arange(100).reshape(4, 5, 5)
    x = np.stack([x_first_channel, x_second_channel], axis=-1)

    expected_y = np.arange(200).reshape(4, 5, 5, 2)
    expected_x_first_channel = np.concatenate(
        ([-40] * 10,
         np.arange(-40, 40),
         [40] * 10)
    ).reshape(4, 5, 5)
    expected_x_second_channel = np.arange(100).reshape(4, 5, 5)
    expected_x = np.stack(
        [expected_x_first_channel, expected_x_second_channel], axis=-1)

    ouput_x, ouput_y = wp.transform(x, y)

    assert ouput_x.shape == expected_x.shape
    assert ouput_y.shape == expected_y.shape
    assert np.all(ouput_x == expected_x)
    assert np.all(ouput_y == expected_y)


def test_keras_image_preprocessor():
    kpx = KerasImagePreprocessorX(scale_down=2, shuffle=False)
    kpy = KerasImagePreprocessorY(scale_down=4, shuffle=False)

    x, y = np.arange(100).reshape(
        4, 5, 5, 1), np.arange(100).reshape(4, 5, 5, 1)

    expected_x = np.arange(100).reshape(4, 5, 5, 1) / 2
    expected_y = np.arange(100).reshape(4, 5, 5, 1) / 4

    output_x, output_y = kpx.transform(x, y)
    output_x, output_y = kpy.transform(output_x, output_y)

    assert expected_x.shape == output_x.shape
    assert expected_y.shape == output_y.shape
    assert np.allclose(expected_x, output_x)
    assert np.all(expected_y == output_y)



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"
__version__ = "0.0.1"


from ..keras.layers import Layer
from ..keras.models import model_from_config
from ..utils import Singleton
from .activations import Activations


class Layers(metaclass=Singleton):
    """
    A singleton that contains all the registered customized layers
    """

    def __init__(self):
        self._layers = {}

    def register(self, key, layer):
        if not issubclass(layer, Layer):
            raise ValueError(
                "The customized layer has to be a subclass"
                + " of keras.layers.Layer"
            )

        if key in self._layers:
            raise KeyError(
                "Duplicated key, please use another key name for this layer"
            )
        else:
            self._layers[key] = layer

    def unregister(self, key):
        if key in self._layers:
            del self._layers[key]

    @property
    def layers(self):
        return self._layers


def register_layer(key, layer):
    """
    Register the customized layer.
    If the key name is already registered, it will raise a KeyError exception

    :param key: the unique key-name of the layer
    :type key: str
    :param layer: the customized layer class
    :type layer: keras.layers.Layer
    """
    Layers().register(key, layer)


def unregister_layer(key):
    """
    Remove the registered layer with the key-name

    :param key: the key-name of the layer to be removed
    :type key: str
    """
    Layers().unregister(key)


def layer_from_config(config):
    if 'class_name' not in config:
        raise ValueError('class_name is needed to define layer')

    if 'config' not in config:
        # auto add empty config for layer with only class_name
        config['config'] = {}
    return model_from_config(
        config,
        custom_objects={**Layers().layers,
                        **Activations().activations})



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"
__version__ = "0.0.1"


from ..keras.layers import Activation
from ..keras.activations import deserialize
from ..utils import Singleton


class Activations(metaclass=Singleton):
    """
    A singleton that contains all the registered customized activations
    """

    def __init__(self):
        self._activations = {}

    def register(self, key, activation):
        if not issubclass(activation, Activation):
            raise ValueError(
                "The customized activation has to be a subclass"
                + " of keras.activations.Activation"
            )

        if key in self._activations:
            raise KeyError(
                "Duplicated key, please use another key for this activation"
            )
        else:
            self._activations[key] = activation

    def unregister(self, key):
        if key in self._activations:
            del self._activations[key]

    @property
    def activations(self):
        return self._activations


def register_activation(key, activation):
    """
    Register the customized activation.
    If the key name is already registered, it will raise a KeyError exception

    :param key: the unique key-name of the activation
    :type key: str
    :param activation: the customized activation class
    :type activation: keras.activations.Activation
    """
    Activations().register(key, activation)


def unregister_activation(key):
    """
    Remove the registered activation with the key-name

    :param key: the key-name of the activation to be removed
    :type key: str
    """
    Activations().unregister(key)


def activation_from_config(config):
    if 'class_name' not in config:
        raise ValueError('class_name is needed to define activation')

    if 'config' not in config:
        # auto add empty config for activation with only class_name
        config['config'] = {}
    return deserialize(config, custom_objects=Activations().activations)



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"
__version__ = "0.0.1"


import pytest
from deoxys.keras.optimizers import Optimizer
from deoxys.model.optimizers import Optimizers
from deoxys.customize import register_optimizer, \
    unregister_optimizer, custom_optimizer
from deoxys.utils import Singleton


@pytest.fixture(autouse=True)
def clear_singleton():
    Singleton._instances = {}  # clear singleton


@pytest.fixture
def optimizer_class():
    class TestOptimizer(Optimizer):
        pass

    yield TestOptimizer


def test_is_singleton():
    optimizers_instance = Optimizers()
    another_instance = Optimizers()

    assert optimizers_instance is another_instance


def test_register_random_obj():
    with pytest.raises(ValueError):
        register_optimizer("TestOptimizer", object)


def test_register_optimizer_success(optimizer_class):
    register_optimizer("TestOptimizer", optimizer_class)

    assert Optimizers()._optimizers["TestOptimizer"] is optimizer_class


def test_register_duplicate_optimizer(optimizer_class):
    register_optimizer("TestOptimizer", optimizer_class)

    with pytest.raises(KeyError):
        register_optimizer("TestOptimizer", optimizer_class)


def test_unregister_optimizer(optimizer_class):
    register_optimizer("TestOptimizer", optimizer_class)

    assert Optimizers()._optimizers["TestOptimizer"] is optimizer_class
    unregister_optimizer("TestOptimizer")

    assert "TestOptimizer" not in Optimizers()._optimizers


def test_decorator():
    @custom_optimizer
    class TestOptimizer2(Optimizer):
        pass

    assert Optimizers()._optimizers["TestOptimizer2"] is TestOptimizer2



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"
__version__ = "0.0.1"


from ..utils import Singleton, is_keras_standalone
from ..keras import backend as K
from ..keras.metrics import Metric, deserialize

if is_keras_standalone():
    from keras.metrics import _ConfusionMatrixConditionCount
    from keras.utils.metrics_utils \
        import update_confusion_matrix_variables, parse_init_thresholds, \
        ConfusionMatrix
else:
    from tensorflow.python.keras.metrics import _ConfusionMatrixConditionCount
    from tensorflow.python.keras.utils.metrics_utils \
        import update_confusion_matrix_variables, parse_init_thresholds, \
        ConfusionMatrix


class Fbeta(Metric):
    def __init__(self, threshold=None, name='Fbeta', dtype=None, beta=1):
        super().__init__(name=name, dtype=dtype)

        self.threshold = 0.5 if threshold is None else threshold
        self.beta = beta

        self.total = self.add_weight(
            'total', initializer='zeros')
        self.count = self.add_weight(
            'count', initializer='zeros')

    def update_state(self, y_true, y_pred, sample_weight=None):
        size = len(y_pred.get_shape().as_list())
        reduce_ax = list(range(1, size))
        eps = 1e-8

        true_positive = K.sum(y_pred * y_true, axis=reduce_ax)
        target_positive = K.sum(K.square(y_true), axis=reduce_ax)
        predicted_positive = K.sum(
            K.square(y_pred), axis=reduce_ax)

        fb_numerator = (1 + self.beta ** 2) * true_positive + eps
        fb_denominator = (
            (self.beta ** 2) * target_positive + predicted_positive + eps
        )
        if sample_weight:
            weight = K.cast(sample_weight, self.dtype)
            total_ops = K.update_add(
                self.total,
                K.sum(weight * fb_numerator / fb_denominator))
        else:
            total_ops = K.update_add(
                self.total, K.sum(fb_numerator / fb_denominator))

        count = K.sum(weight) if sample_weight else K.cast(
            K.shape(y_pred)[0], y_pred.dtype)

        count_ops = K.update_add(self.count, count)

        if is_keras_standalone():
            return [total_ops, count_ops]

    def result(self):
        return self.total / self.count

    def get_config(self):
        config = {'threshold': self.threshold,
                  'beta': self.beta}
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))


class BinaryFbeta(_ConfusionMatrixConditionCount):
    """
    Calculate the micro f1 score in the set of data
    """

    def __init__(self,
                 thresholds=None,
                 name='BinaryFbeta',
                 dtype=None, beta=1):
        super(Metric, self).__init__(name=name, dtype=dtype)

        self.init_thresholds = thresholds
        self.thresholds = parse_init_thresholds(
            thresholds, default_threshold=0.5)
        num_thresholds = len(self.thresholds)
        self.true_positives = self.add_weight(
            'true_positives',
            shape=(num_thresholds,),
            initializer='zeros')
        self.true_negatives = self.add_weight(
            'true_negatives',
            shape=(num_thresholds,),
            initializer='zeros')
        self.false_positives = self.add_weight(
            'false_positives',
            shape=(num_thresholds,),
            initializer='zeros')
        self.false_negatives = self.add_weight(
            'false_negatives',
            shape=(num_thresholds,),
            initializer='zeros')
        self.beta = beta

    def update_state(self, y_true, y_pred, sample_weight=None):
        # https://github.com/tensorflow/tensorflow/issues/30711
        # Remove return statement in case tensorflow.keras
        update = update_confusion_matrix_variables(
            {ConfusionMatrix.TRUE_POSITIVES: self.true_positives,
             ConfusionMatrix.TRUE_NEGATIVES: self.true_negatives,
             ConfusionMatrix.FALSE_POSITIVES: self.false_positives,
             ConfusionMatrix.FALSE_NEGATIVES: self.false_negatives},
            y_true,
            y_pred,
            thresholds=self.thresholds,
            sample_weight=sample_weight)

        if is_keras_standalone():
            return update

    def result(self):
        res = []
        eps = 1e-8
        for i in range(len(self.thresholds)):
            fb_numerator = (1 + self.beta ** 2) * self.true_positives + eps
            fb_denominator = (
                (self.beta ** 2) * (self.true_positives + self.false_negatives)
                + self.true_positives + self.false_positives + eps
            )
            fscore = fb_numerator / fb_denominator
            res.append(fscore)
        if len(res) == 1:
            return res[0]
        return res

    def get_config(self):
        config = {'thresholds': self.init_thresholds}
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))


class Metrics(metaclass=Singleton):
    """
    A singleton that contains all the registered customized metrics
    """

    def __init__(self):
        self._metrics = {
            'BinaryFbeta': BinaryFbeta,
            'Fbeta': Fbeta
        }

    def register(self, key, metric):
        if not issubclass(metric, Metric):
            raise ValueError(
                "The customized metric has to be a subclass"
                + " of keras.metrics.Metric"
            )

        if key in self._metrics:
            raise KeyError(
                "Duplicated key, please use another key for this metric"
            )
        else:
            self._metrics[key] = metric

    def unregister(self, key):
        if key in self._metrics:
            del self._metrics[key]

    @property
    def metrics(self):
        return self._metrics


def register_metric(key, metric):
    """
    Register the customized metric.
    If the key name is already registered, it will raise a KeyError exception

    :param key: the unique key-name of the metric
    :type key: str
    :param metric: the customized metric class
    :type metric: keras.metrics.Metric
    """
    Metrics().register(key, metric)


def unregister_metric(key):
    """
    Remove the registered metric with the key-name

    :param key: the key-name of the metric to be removed
    :type key: str
    """
    Metrics().unregister(key)


def metric_from_config(config):
    if type(config) == dict:
        if 'class_name' not in config:
            raise ValueError('class_name is needed to define metric')

        if 'config' not in config:
            # auto add empty config for metric with only class_name
            config['config'] = {}
        return deserialize(
            config,
            custom_objects=Metrics().metrics)

    return deserialize(config, custom_objects=Metrics().metrics)



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"
__version__ = "0.0.1"

from ..keras.optimizers import Optimizer, deserialize
from ..utils import Singleton


class Optimizers(metaclass=Singleton):
    """
    A singleton that contains all the registered customized optimizers
    """

    def __init__(self):
        self._optimizers = {}

    def register(self, key, optimizer):
        if not issubclass(optimizer, Optimizer):
            raise ValueError(
                "The customized optimizer has to be a subclass"
                + " of keras.optimizers.Optimizer"
            )

        if key in self._optimizers:
            raise KeyError(
                "Duplicated key, please use another key for this optimizer"
            )
        else:
            self._optimizers[key] = optimizer

    def unregister(self, key):
        if key in self._optimizers:
            del self._optimizers[key]

    @property
    def optimizers(self):
        return self._optimizers


def register_optimizer(key, optimizer):
    """
    Register the customized optimizer.
    If the key name is already registered, it will raise a KeyError exception

    :param key: the unique key-name of the optimizer
    :type key: str
    :param optimizer: the customized optimizer class
    :type optimizer: keras.optimizers.Optimizer
    """
    Optimizers().register(key, optimizer)


def unregister_optimizer(key):
    """
    Remove the registered optimizer with the key-name

    :param key: the key-name of the optimizer to be removed
    :type key: str
    """
    Optimizers().unregister(key)


def optimizer_from_config(config):
    if 'class_name' not in config:
        raise ValueError('class_name is needed to define optimizer')

    if 'config' not in config:
        # auto add empty config for optimizer with only class_name
        config['config'] = {}
    return deserialize(config, custom_objects=Optimizers().optimizers)



import json
from . import read_file


def load_json_config(*args):
    data = []
    for arg in args:
        if arg is None:
            data.append(arg)
        elif type(arg) == dict:
            data.append(arg)
        elif type(arg) == str:
            try:
                if arg.startswith('{') or arg.startswith('['):
                    data.append(json.loads(arg))
                else:
                    data.append(json.loads(read_file(arg)))
            except json.JSONDecodeError:
                data.append(None)
                raise Warning('Decode JSON failed. Return None instead.')
            except IOError:
                data.append(None)
                raise Warning('File not found. Return None instead.')
            except Exception as ex:
                data.append(None)
                raise Warning('An error occur. {}'.format(ex))
        else:
            data.append(None)
            raise Warning('Invalid datatype. Return None instead.')

    if len(data) == 1:
        return data[0]
    return data



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"
__version__ = "0.0.1"


import numpy as np


class DataGenerator:
    def __init__(self):
        pass

    @property
    def total_batch(self):
        """
        Total number of batches to iterate all data

        : return: Total number of batches to iterate all data
        : rtype: int
        """
        return 0

    def generate(self):
        raise NotImplementedError()


class HDF5DataGenerator(DataGenerator):
    def __init__(self, h5file, batch_size=32, batch_cache=10,
                 preprocessors=None,
                 x_name='x', y_name='y', folds=None):
        if not folds or not h5file:
            raise ValueError("h5file or folds is empty")

        # Checking for existence of folds and dataset
        group_names = h5file.keys()
        dataset_names = []
        str_folds = [str(fold) for fold in folds]
        for fold in str_folds:
            if fold not in group_names:
                raise RuntimeError(
                    'HDF5 file: Fold name "{0}" is not in this h5 file'
                    .format(fold))
            if dataset_names:
                if h5file[fold].keys() != dataset_names:
                    raise RuntimeError(
                        'HDF5 file: All folds should have the same structure')
            else:
                dataset_names = h5file[fold].keys()
                if x_name not in dataset_names or y_name not in dataset_names:
                    raise RuntimeError(
                        'HDF5 file: {0} or {1} is not in the file'
                        .format(x_name, y_name))

        # Checking for valida preprocessor
        if preprocessors:
            if type(preprocessors) == list:
                for pp in preprocessors:
                    if not callable(getattr(pp, 'transform', None)):
                        raise ValueError(
                            'Preprocessor should have a "transform" method')
            else:
                if not callable(getattr(preprocessors, 'transform', None)):
                    raise ValueError(
                        'Preprocessor should have a "transform" method')

        self.hf = h5file
        self.batch_size = batch_size
        self.seg_size = batch_size * batch_cache
        self.preprocessors = preprocessors
        self.x_name = x_name
        self.y_name = y_name

        self.folds = str_folds

        # Cache first segment for first fold
        self.index = 0
        self.seg_index = 0

        first_fold_name = self.folds[0]

        self.x_cur = self.hf[first_fold_name][self.x_name][:self.seg_size]
        self.y_cur = self.hf[first_fold_name][self.y_name][:self.seg_size]

        if self.preprocessors:
            if type(self.preprocessors) == list:
                for preprocessor in self.preprocessors:
                    self.x_cur, self.y_cur = preprocessor.transform(
                        self.x_cur, self.y_cur)
            else:
                self.x_cur, self.y_cur = self.preprocessors.transform(
                    self.x_cur, self.y_cur)

        # Get the total length of the first fold
        self.fold_len = len(self.hf[first_fold_name][self.y_name])

        self._total_batch = None

    @property
    def total_batch(self):
        """
        Total number of batches to iterate all data.
        It will be used as the number of steps per epochs when training or
        validating data in a model.

        : return: Total number of batches to iterate all data
        : rtype: int
        """
        if self._total_batch is None:
            total_batch = 0
            fold_names = self.folds

            for fold_name in fold_names:
                total_batch += np.ceil(
                    len(self.hf[fold_name][self.y_name]) / self.batch_size)
            self._total_batch = int(total_batch)
        return self._total_batch

    def next_fold(self):
        # Reset segment index
        self.seg_index = 0

        # Remove previous fold index and move to next one
        self.folds.append(self.folds.pop(0))

        fold_name = self.folds[0]
        y = self.hf[fold_name][self.y_name]

        # Recalculate the total length
        self.fold_len = len(y)

    def next_seg(self):
        # Reset index
        self.index = 0

        # Move segment index
        self.seg_index += self.seg_size

        # When all segments fold has been yielded, move to next fold
        if self.seg_index >= self.fold_len:
            self.next_fold()

        # store local variable after seg_index changed
        seg_index = self.seg_index

        fold_name = self.folds[0]
        # The last segment may has less items than seg_size
        if seg_index + self.seg_size >= self.fold_len:
            self.x_cur = self.hf[fold_name][self.x_name][seg_index:]
            self.y_cur = self.hf[fold_name][self.y_name][seg_index:]
        else:
            next_seg_index = seg_index + self.seg_size

            self.x_cur = self.hf[fold_name][
                self.x_name][seg_index:next_seg_index]
            self.y_cur = self.hf[fold_name][
                self.y_name][seg_index:next_seg_index]

        # Apply preprocessor
        if self.preprocessors:
            if type(self.preprocessors) == list:
                for preprocessor in self.preprocessors:
                    self.x_cur, self.y_cur = preprocessor.transform(
                        self.x_cur, self.y_cur)
            else:
                self.x_cur, self.y_cur = self.preprocessors.transform(
                    self.x_cur, self.y_cur)

    def generate(self):
        """
        Create a generator that generate a batch of data

        :yield: batch of (input, target)
        :rtype: tuple of 2 arrays
        """
        while True:
            # When all batches of data are yielded, move to next seg
            if self.index >= self.seg_size or \
                    self.seg_index + self.index >= self.fold_len:
                self.next_seg()

            # Index may has been reset. Thus, call after next_seg
            index = self.index

            # The last batch of data may not have less than batch_size items
            if index + self.batch_size >= self.seg_size or \
                    self.seg_index + index + self.batch_size >= self.fold_len:
                batch_x = self.x_cur[index:]
                batch_y = self.y_cur[index:]
            else:
                # Take the next batch
                batch_x = self.x_cur[index:(index + self.batch_size)]
                batch_y = self.y_cur[index:(index + self.batch_size)]

            self.index += self.batch_size
            yield batch_x, batch_y



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"
__version__ = "0.0.1"


from deoxys.keras.preprocessing import ImageDataGenerator
from ..utils import Singleton


class BasePreprocessor:
    def __init__(self, **kwargs):
        """
        Wrapper for creating a trained/fitted Preprocessor instance
        """
        pass

    def transform(self, x, y):
        return x, y


class DummyPreprocessor:
    def __init__(self):
        pass

    def transform(self, x, y):
        return x, y


class SingleChannelPreprocessor(BasePreprocessor):
    def transform(self, x, y):
        x_shape = tuple(list(x.shape) + [1])
        y_shape = tuple(list(y.shape) + [1])

        return x.reshape(x_shape), y.reshape(y_shape)


class WindowingPreprocessor(BasePreprocessor):
    """Used to set the dynamic range of an image.
    """

    def __init__(self, window_center, window_width, channel):
        self.window_center, self.window_width = window_center, window_width
        self.channel = channel

    def perform_windowing(self, image):
        image = image - self.window_center
        image[image < -self.window_width / 2] = -self.window_width / 2
        image[image > self.window_width / 2] = self.window_width / 2
        return image

    def transform(self, images, targets):
        images = images.copy()
        images[..., self.channel] = self.perform_windowing(
            images[..., self.channel])
        return images, targets


class KerasImagePreprocessorX(BasePreprocessor):
    def __init__(self,
                 shuffle=False,
                 featurewise_center=False,
                 samplewise_center=False,
                 featurewise_std_normalization=False,
                 samplewise_std_normalization=False,
                 zca_whitening=False, zca_epsilon=1e-06,
                 rotation_range=0, width_shift_range=0.0,
                 height_shift_range=0.0,
                 brightness_range=None, shear_range=0.0, zoom_range=0.0,
                 channel_shift_range=0.0, fill_mode='nearest', cval=0.0,
                 horizontal_flip=False, vertical_flip=False, rescale=None,
                 scale_down=None,
                 preprocessing_function=None, data_format='channels_last',
                 interpolation_order=1, dtype='float32'):

        self.shuffle = shuffle

        if scale_down and (rescale is None):
            rescale = 1 / scale_down

        self.preprocessor = ImageDataGenerator(
            featurewise_center=featurewise_center,
            samplewise_center=samplewise_center,
            featurewise_std_normalization=featurewise_std_normalization,
            samplewise_std_normalization=samplewise_std_normalization,
            zca_whitening=zca_whitening, zca_epsilon=zca_epsilon,
            rotation_range=rotation_range, width_shift_range=width_shift_range,
            height_shift_range=height_shift_range,
            brightness_range=brightness_range, shear_range=shear_range,
            zoom_range=zoom_range,
            channel_shift_range=channel_shift_range, fill_mode=fill_mode,
            cval=cval,
            horizontal_flip=horizontal_flip, vertical_flip=vertical_flip,
            rescale=rescale,
            preprocessing_function=preprocessing_function,
            data_format=data_format, dtype=dtype)

    def transform(self, x, y):
        return next(self.preprocessor.flow(x, batch_size=x.shape[0],
                                           shuffle=self.shuffle)), y


class KerasImagePreprocessorY(BasePreprocessor):
    def __init__(self,
                 shuffle=True,
                 featurewise_center=False,
                 samplewise_center=False,
                 featurewise_std_normalization=False,
                 samplewise_std_normalization=False,
                 zca_whitening=False, zca_epsilon=1e-06,
                 rotation_range=0, width_shift_range=0.0,
                 height_shift_range=0.0,
                 brightness_range=None, shear_range=0.0, zoom_range=0.0,
                 channel_shift_range=0.0, fill_mode='nearest', cval=0.0,
                 horizontal_flip=False, vertical_flip=False, rescale=None,
                 scale_down=None,
                 preprocessing_function=None, data_format='channels_last',
                 dtype='float32'):
        self.shuffle = shuffle

        if scale_down and (rescale is None):
            rescale = 1 / scale_down

        self.preprocessor = ImageDataGenerator(
            featurewise_center=featurewise_center,
            samplewise_center=samplewise_center,
            featurewise_std_normalization=featurewise_std_normalization,
            samplewise_std_normalization=samplewise_std_normalization,
            zca_whitening=zca_whitening, zca_epsilon=zca_epsilon,
            rotation_range=rotation_range, width_shift_range=width_shift_range,
            height_shift_range=height_shift_range,
            brightness_range=brightness_range, shear_range=shear_range,
            zoom_range=zoom_range,
            channel_shift_range=channel_shift_range, fill_mode=fill_mode,
            cval=cval,
            horizontal_flip=horizontal_flip, vertical_flip=vertical_flip,
            rescale=rescale,
            preprocessing_function=preprocessing_function,
            data_format=data_format,
            dtype=dtype)

    def transform(self, x, y):
        return x, next(self.preprocessor.flow(y, batch_size=x.shape[0],
                                              shuffle=self.shuffle))


class Preprocessors(metaclass=Singleton):
    """
    A singleton that contains all the registered customized preprocessors
    """

    def __init__(self):
        self._preprocessors = {
            'WindowingPreprocessor': WindowingPreprocessor,
            'SingleChannelPreprocessor': SingleChannelPreprocessor,
            'KerasImagePreprocessorX': KerasImagePreprocessorX,
            'KerasImagePreprocessorY': KerasImagePreprocessorY
        }

    def register(self, key, preprocessor):
        if not issubclass(preprocessor, BasePreprocessor):
            raise ValueError(
                "The customized preprocessor has to be a subclass"
                + " of deoxys.data.BasePreprocessor"
            )

        if key in self._preprocessors:
            raise KeyError(
                "Duplicated key, please use another key for this preprocessor"
            )
        else:
            self._preprocessors[key] = preprocessor

    def unregister(self, key):
        if key in self._preprocessors:
            del self._preprocessors[key]

    @property
    def preprocessors(self):
        return self._preprocessors


def register_preprocessor(key, preprocessor):
    """
    Register the customized preprocessor.
    If the key name is already registered, it will raise a KeyError exception

    :param key: the unique key-name of the preprocessor
    :type key: str
    :param preprocessor: the customized preprocessor class
    :type preprocessor: deoxys.data.BasePreprocessor
    """
    Preprocessors().register(key, preprocessor)


def unregister_preprocessor(key):
    """
    Remove the registered preprocessor with the key-name

    :param key: the key-name of the preprocessor to be removed
    :type key: str
    """
    Preprocessors().unregister(key)


def _deserialize(config, custom_objects={}):
    predefined_obj = {
        'DummyPreprocessor': DummyPreprocessor
    }

    predefined_obj.update(custom_objects)

    return predefined_obj[config['class_name']](**config['config'])


def preprocessor_from_config(config):
    if 'class_name' not in config:
        raise ValueError('class_name is needed to define preprocessor')

    if 'config' not in config:
        # auto add empty config for preprocessor with only class_name
        config['config'] = {}
    return _deserialize(config, custom_objects=Preprocessors().preprocessors)



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"
__version__ = "0.0.1"


"""
This file contains multiple helper function for plotting diagram and images
using matplotlib
"""


import pandas as pd
import matplotlib.pyplot as plt
from .data_utils import read_csv


def plot_log_performance_from_csv(filepath, output_path):
    """
    Plot and save multiple performance figure using a log file generated from
    tensorflow.keras.callbacks.CSVLogger

    :param filepath: filename of the log file
    :type filepath: str
    :param output_path: path to the folder for saving plotted diagram
    :type output_path: str
    """
    df = read_csv(filepath, index_col='epoch')

    # Plot all data
    _plot_data(df, 'All parameters', df.columns, output_path + '/all.png')

    # Plot train data
    train_keys = [key for key in df.columns if 'val' not in key]
    _plot_data(df, 'Train Performance', train_keys, output_path + '/train.png')

    # Plot val data
    val_keys = [key for key in df.columns if 'val' in key]
    _plot_data(df, 'Validation Performance',
               val_keys, output_path + '/val.png')

    # Plot compare train and val
    compare = [[key, 'val_' + key] for key in train_keys]
    for keys in compare:
        _plot_data(df, '{} Performance'.format(
            keys[0]), keys, output_path + '/{}.png'.format(keys[0]))


def plot_evaluation_performance_from_csv(filepath, output_path):
    # Load data to file
    df = read_csv(filepath, index_col='epoch')

    # Plot evaluation
    _plot_data(df, 'Evaluation Performance', df.columns,
               output_path + '/evaluation.png')


def mask_prediction(output_path, image, true_mask, pred_mask,
                    title='Predicted',
                    mask_levels=None, channel=None):
    """
    Generate and save predicted images with true mask and predicted mask as
    contouring lines

    :param output_path: path to folder for saving the output images
    :type output_path: str
    :param image: a collection of original 2D image data
    :type image: numpy array / collection
    :param true_mask: a collection of true mask data
    :type true_mask: numpy array / collection
    :param pred_mask: a collection of predicted mask data
    :type pred_mask: numpy array / collection
    :param title: title of the diagram, defaults to 'Predicted'
    :type title: str, optional
    :param mask_levels: mask_levels when contouring the images,
    defaults to None
    :type mask_levels: int, or list of int or float, optional
    :param channel: if the original image has multiple channels, this indicates
    which channel to plot the images, defaults to None
    :type channel: int, optional
    """
    if not mask_levels:
        mask_levels = 1
    kwargs = {}
    if not channel:
        if (len(image.shape) == 2
                or (len(image.shape) == 3 and image.shape[2] == 3)):
            image_data = image
        else:
            image_data = image[..., 0]
            kwargs['cmap'] = 'gray'
    else:
        image_data = image[..., channel]
        kwargs['cmap'] = 'gray'

    true_mask_data = true_mask
    pred_mask_data = pred_mask

    if len(true_mask_data.shape) == 3:
        true_mask_data = true_mask[..., 0]
        pred_mask_data = pred_mask[..., 0]

    plt.figure()
    plt.imshow(image_data, **kwargs)
    true_con = plt.contour(
        true_mask_data, 1, levels=mask_levels, colors='yellow')
    pred_con = plt.contour(
        pred_mask_data, 1, levels=mask_levels, colors='red')

    plt.title(title)
    plt.legend([true_con.collections[0],
                pred_con.collections[0]], ['True', 'Predicted'])
    plt.savefig(output_path)
    plt.close('all')


def plot_images_w_predictions(output_path, image, true_mask, pred_mask,
                              title='Predicted',
                              channel=None):
    """
    Generate and save predicted images with true mask and predicted mask as
    separate images

    :param output_path: path to folder for saving the output images
    :type output_path: str
    :param image: a collection of original 2D image data
    :type image: numpy array / collection
    :param true_mask: a collection of true mask data
    :type true_mask: numpy array / collection
    :param pred_mask: a collection of predicted mask data
    :type pred_mask: numpy array / collection
    :param title: title of the diagram, defaults to 'Predicted'
    :type title: str, optional
    :param channel: if the original image has multiple channels, this indicates
    which channel to plot the images, defaults to None
    :type channel: int, optional
    """
    kwargs = {}
    if not channel:
        if (len(image.shape) == 2
                or (len(image.shape) == 3 and image.shape[2] == 3)):
            image_data = image
        else:
            image_data = image[..., 0]
            kwargs['cmap'] = 'gray'
    else:
        image_data = image[..., channel]
        kwargs['cmap'] = 'gray'

    true_mask_data = true_mask
    pred_mask_data = pred_mask

    if len(true_mask_data.shape) == 3:
        true_mask_data = true_mask[..., 0]
        pred_mask_data = pred_mask[..., 0]

    fig, (img_ax, true_ax, pred_ax) = plt.subplots(1, 3)
    img_ax.imshow(image_data, **kwargs)
    img_ax.set_title('Images')
    true_ax.imshow(true_mask_data)
    true_ax.set_title('True Mask')
    pred_ax.imshow(pred_mask_data)
    pred_ax.set_title('Predicted Mask')

    plt.suptitle(title)
    plt.savefig(output_path)
    plt.close('all')


def _plot_data(dataframe, name, columns, filename):
    ax = dataframe[columns].plot()
    epoch_num = dataframe.shape[0]
    if epoch_num < 20:
        ax.set_xticks(dataframe.index)
    else:
        tick_distance = epoch_num // 10
        ax.set_xticks([tick for i, tick in enumerate(
            dataframe.index) if i % tick_distance == 0])
    ax.set_title(name)
    plt.savefig(filename)



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"
__version__ = "0.0.1"


class Singleton(type):
    """
    A meta class to create a singleton in this project.

    Example:
    ```
    class SingletonClassName(metaclass=Singleton):
        # attributes and method here
    ```
    """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(
                *args, **kwargs
            )

        return cls._instances[cls]



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"
__version__ = "0.0.1"


from .model import *
from . import layers, activations, callbacks, metrics, optimizers



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"
__version__ = "0.0.1"


from ..model.layers import register_layer, unregister_layer
from ..model.activations import register_activation, unregister_activation
from ..model.callbacks import register_callback, unregister_callback
from ..model.losses import register_loss, unregister_loss
from ..model.metrics import register_metric, unregister_metric
from ..model.optimizers import register_optimizer, unregister_optimizer
from ..data.data_reader import register_datareader, unregister_datareader
from ..data.preprocessor import register_preprocessor, unregister_preprocessor
from ..loaders.architecture import register_architecture


def custom_architecture(class_def):
    register_architecture(class_def.__name__, class_def)
    return class_def


def custom_layer(class_def):
    register_layer(class_def.__name__, class_def)
    return class_def


def custom_activation(class_def):
    register_activation(class_def.__name__, class_def)
    return class_def


def custom_callback(class_def):
    register_callback(class_def.__name__, class_def)
    return class_def


def custom_loss(class_def):
    register_loss(class_def.__name__, class_def)
    return class_def


def custom_metric(class_def):
    register_metric(class_def.__name__, class_def)
    return class_def


def custom_optimizer(class_def):
    register_optimizer(class_def.__name__, class_def)
    return class_def


def custom_datareader(class_def):
    register_datareader(class_def.__name__, class_def)
    return class_def


def custom_preprocessor(class_def):
    register_preprocessor(class_def.__name__, class_def)
    return class_def



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"
__version__ = "0.0.1"


from ..database import Tables, SessionStatus, SessionAttr, ExperimentAttr, \
    RefAttr, ConfigRef, LogAttr
from itertools import product
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from . import ExperimentDB


class MultiExperimentDB:

    def __init__(self, dbclient):
        self.dbclient = dbclient

    @property
    def experiments(self):
        """
        List all experiements

        :return: [description]
        :rtype: [type]
        """
        all_exp = self.dbclient.find_all(Tables.EXPERIMENTS)

        return self.dbclient.to_pandas(all_exp)

    def sessions_from_experiments(self, experiment_id):
        """
        List all sessions (created, training, finished, failed) of
        an experiments

        :param experiment_id: [description]
        :type experiment_id: [type]
        :return: [description]
        :rtype: [type]
        """
        sessions = self.dbclient.find_by_col(
            Tables.SESSIONS, SessionAttr.EXPERIMENT_ID,
            self.dbclient.to_fk(experiment_id))

        return self.dbclient.to_pandas(sessions)

    def run_new_session(self, experiment_id, epochs, log_base_path='logs',
                        model_checkpoint_period=2,
                        prediction_checkpoint_period=2):
        """
        Start a new session of an experiment

        :param experiment_id: [description]
        :type experiment_id: [type]
        :param epochs: [description]
        :type epochs: [type]
        :param log_base_path: [description], defaults to 'logs'
        :type log_base_path: str, optional
        :param model_checkpoint_period: [description], defaults to 2
        :type model_checkpoint_period: int, optional
        :param prediction_checkpoint_period: [description], defaults to 2
        :type prediction_checkpoint_period: int, optional
        :return: [description]
        :rtype: [type]
        """
        exp = ExperimentDB(
            self.dbclient, experiment_id=experiment_id,
            log_base_path=log_base_path
        ).run_experiment(
            model_checkpoint_period=model_checkpoint_period,
            prediction_checkpoint_period=prediction_checkpoint_period,
            save_origin_images=False, verbose=1, epochs=epochs)

        return exp

    def run_multiple_new_session(self, num, experiment_id, epochs,
                                 log_base_path='logs',
                                 model_checkpoint_period=2,
                                 prediction_checkpoint_period=2):
        """
        Run `num` number of new sessions of an experiment

        :param num: [description]
        :type num: [type]
        :param experiment_id: [description]
        :type experiment_id: [type]
        :param epochs: [description]
        :type epochs: [type]
        :param log_base_path: [description], defaults to 'logs'
        :type log_base_path: str, optional
        :param model_checkpoint_period: [description], defaults to 2
        :type model_checkpoint_period: int, optional
        :param prediction_checkpoint_period: [description], defaults to 2
        :type prediction_checkpoint_period: int, optional
        """
        return_exps = []
        for _ in range(num):
            try:
                exp = ExperimentDB(
                    self.dbclient, experiment_id=experiment_id,
                    log_base_path=log_base_path
                ).run_experiment(
                    model_checkpoint_period=model_checkpoint_period,
                    prediction_checkpoint_period=prediction_checkpoint_period,
                    save_origin_images=False, verbose=1, epochs=epochs)
                return_exps.append(exp)
            except Exception:
                pass

        return return_exps

    def continue_session(self, session_id, epochs, log_base_path='logs',
                         model_checkpoint_period=2,
                         prediction_checkpoint_period=2):
        """
        Continue a session from a model checkpoint

        :param session_id: [description]
        :type session_id: [type]
        :param epochs: [description]
        :type epochs: [type]
        :param log_base_path: [description], defaults to 'logs'
        :type log_base_path: str, optional
        :param model_checkpoint_period: [description], defaults to 2
        :type model_checkpoint_period: int, optional
        :param prediction_checkpoint_period: [description], defaults to 2
        :type prediction_checkpoint_period: int, optional
        :return: [description]
        :rtype: [type]
        """
        exp = ExperimentDB(
            self.dbclient, session_id=session_id, log_base_path=log_base_path
        ).run_experiment(
            model_checkpoint_period=model_checkpoint_period,
            prediction_checkpoint_period=prediction_checkpoint_period,
            save_origin_images=False, verbose=1, epochs=epochs)

        return exp

    def continue_multiple_session(self, session_id_list, epochs,
                                  log_base_path='logs',
                                  model_checkpoint_period=2,
                                  prediction_checkpoint_period=2):
        """
        Continue multiple sessions

        :param session_id_list: [description]
        :type session_id_list: [type]
        :param epochs: [description]
        :type epochs: [type]
        :param log_base_path: [description], defaults to 'logs'
        :type log_base_path: str, optional
        :param model_checkpoint_period: [description], defaults to 2
        :type model_checkpoint_period: int, optional
        :param prediction_checkpoint_period: [description], defaults to 2
        :type prediction_checkpoint_period: int, optional
        """
        return_exps = []
        for session_id in session_id_list:
            try:
                exp = ExperimentDB(
                    self.dbclient, session_id=session_id,
                    log_base_path=log_base_path
                ).run_experiment(
                    model_checkpoint_period=model_checkpoint_period,
                    prediction_checkpoint_period=prediction_checkpoint_period,
                    save_origin_images=False, verbose=1, epochs=epochs)

                return_exps.append(exp)
            except Exception:
                pass

        return return_exps

    def session_performance(self, session_id, metrics=None,
                            ax=None, shading='std'):
        perf = self.dbclient.to_pandas(
            self.dbclient.find_by_col(
                Tables.LOGS, LogAttr.SESSION_ID,
                self.dbclient.to_fk(session_id)))

        not_include = ['_id', LogAttr.SESSION_ID, LogAttr.EPOCH]
        metric_list = metrics or [
            col for col in perf.columns if col not in not_include]
        perf = perf.groupby(LogAttr.EPOCH).agg({
            metric: ['mean', lambda val: np.mean(
                val) - np.std(val), lambda val: np.mean(val) + np.std(val)]
            for metric in metric_list
        })

        postfixes = ['', '_nev', '_pos']
        perf.columns = ['{}{}'.format(metric, postfix)
                        for metric in metric_list for postfix in postfixes]

        if type(session_id) == list:
            # plot mean data instead
            pass
        if ax is None:
            ax = plt.axes()
        epochs = perf.index
        for metric in metric_list:
            ax.plot(epochs, perf[metric], label=metric)
            ax.fill_between(epochs, perf[metric + '_nev'],
                            perf[metric + '_pos'], alpha=0.2)

        ax.legend()
        return ax

    def experiment_performance(self, experiment_id, metrics=None, ax=None,
                               shading='std'):
        sessions = self.dbclient.to_pandas(
            self.dbclient.find_by_col(
                Tables.SESSIONS, SessionAttr.EXPERIMENT_ID,
                self.dbclient.to_fk(experiment_id)))

        return self.session_performance(
            list(sessions['_id'].values), metrics, ax, shading)

    def new_experiment_from_full_config(self, name, config, description=''):
        """
        Add a new experiement from a configuration JSON

        :param name: [description]
        :type name: [type]
        :param config: [description]
        :type config: [type]
        :param description: [description], defaults to ''
        :type description: str, optional
        :return: [description]
        :rtype: [type]
        """
        new_exp = {
            ExperimentAttr.NAME: name,
            ExperimentAttr.CONFIG: config,
            ExperimentAttr.DESC: description
        }

        return self.dbclient.insert(Tables.EXPERIMENTS, new_exp).inserted_id

    def new_experiment_from_h5file(self, name, file_path, description=''):
        """
        Add a new experiment from a h5 model file.
        This is used in case we want to train a modified model on disk.

        :param name: [description]
        :type name: [type]
        :param file_path: [description]
        :type file_path: [type]
        :param description: [description], defaults to ''
        :type description: str, optional
        :return: [description]
        :rtype: [type]
        """
        new_exp = {
            ExperimentAttr.NAME: name,
            ExperimentAttr.SAVED_MODEL_LOC: file_path,
            ExperimentAttr.DESC: description
        }

        return self.dbclient.insert(Tables.EXPERIMENTS, new_exp).inserted_id

    def new_experiment_from_components(self, dataset_params, input_params,
                                       architecture, model_params,
                                       train_params=None,
                                       name='', description=''):
        """
        Create a new experiment by combining pre-defined JSON config

        :param dataset_params: [description]
        :type dataset_params: [type]
        :param input_params: [description]
        :type input_params: [type]
        :param architecture: [description]
        :type architecture: [type]
        :param model_params: [description]
        :type model_params: [type]
        :param train_params: [description], defaults to None
        :type train_params: [type], optional
        :param name: [description], defaults to ''
        :type name: str, optional
        :param description: [description], defaults to ''
        :type description: str, optional
        :return: [description]
        :rtype: [type]
        """
        exp_name = name or ''
        try:
            # find each component by id then create the full JSON
            dataset_params_obj = self.dbclient.find_by_id(
                ConfigRef.DATASET_PARAMS, dataset_params)
            dataset_params_cf = dataset_params_obj[RefAttr.CONFIG]

            input_params_obj = self.dbclient.find_by_id(
                ConfigRef.INPUT_PARAMS, input_params)
            input_params_cf = input_params_obj[RefAttr.CONFIG]

            architecture_obj = self.dbclient.find_by_id(
                ConfigRef.ARCHITECTURE, architecture)
            architecture_cf = architecture_obj[RefAttr.CONFIG]

            model_params_obj = self.dbclient.find_by_id(
                ConfigRef.MODEL_PARAMS, model_params)
            model_params_cf = model_params_obj[RefAttr.CONFIG]

            if train_params:
                train_params_obj = self.dbclient.find_by_id(
                    ConfigRef.TRAIN_PARAMS, train_params)
                train_params_cf = train_params_obj[RefAttr.CONFIG]
            else:
                train_params_cf = '{ }'

            config = self._new_config_from_components(
                dataset_params_cf, input_params_cf, architecture_cf,
                model_params_cf, train_params_cf
            )

            if not name:
                exp_name = ' - '.join([
                    dataset_params_obj[RefAttr.NAME],
                    input_params_obj[RefAttr.NAME],
                    architecture_obj[RefAttr.NAME],
                    model_params_obj[RefAttr.NAME],
                    train_params_obj[RefAttr.NAME] if train_params else '',
                ])
            return self.new_experiment_from_full_config(
                exp_name, config, description=description)
        except Exception:
            return None

    def new_multi_experiments_from_components(self, dataset_params,
                                              input_params,
                                              architecture, model_params,
                                              train_params=None):
        """
        Create multiple experiments by combination of all components

        :param dataset_params: [description]
        :type dataset_params: [type]
        :param input_params: [description]
        :type input_params: [type]
        :param architecture: [description]
        :type architecture: [type]
        :param model_params: [description]
        :type model_params: [type]
        :param train_params: [description], defaults to None
        :type train_params: [type], optional
        :param name: [description], defaults to ''
        :type name: str, optional
        :param description: [description], defaults to ''
        :type description: str, optional
        """
        # Assuming they're all list
        components = [dataset_params, input_params, architecture, model_params,
                      train_params if train_params else [None]]
        args_list = product(*components)

        inserted_ids = []
        for args in args_list:
            inserted_ids.append(
                self.new_multi_experiments_from_components(*args))
        return inserted_ids

    def rename_experiment(self, experiment_id, new_name):
        """
        Rename an experiment by id

        :param experiment_id: [description]
        :type experiment_id: [type]
        :param new_name: [description]
        :type new_name: [type]
        """
        return self.dbclient.update_by_id(Tables.EXPERIMENTS, experiment_id, {
            ExperimentAttr.NAME: new_name
        })

    def update_experiment_description(self, experiment_id, description):
        """
        Rename an experiment by id

        :param experiment_id: [description]
        :type experiment_id: [type]
        :param new_name: [description]
        :type new_name: [type]
        """
        return self.dbclient.update_by_id(Tables.EXPERIMENTS, experiment_id, {
            ExperimentAttr.DESC: description
        })

    def new_architecture_config(self, name, config, description=''):
        """
        Add a new JSON config for architecture.

        :param name: [description]
        :type name: [type]
        :param config: [description]
        :type config: [type]
        :param description: [description], defaults to ''
        :type description: str, optional
        :return: [description]
        :rtype: [type]
        """
        return self._new_ref_config(
            ConfigRef.ARCHITECTURE, name, config, description)

    def new_dataset_config(self, name, config, description=''):
        """
        Add a new JSON config for dataset_params

        :param name: [description]
        :type name: [type]
        :param config: [description]
        :type config: [type]
        :param description: [description], defaults to ''
        :type description: str, optional
        :return: [description]
        :rtype: [type]
        """
        return self._new_ref_config(
            ConfigRef.DATASET_PARAMS, name, config, description)

    def new_input_config(self, name, config, description=''):
        """
        Add a new JSON config for input_params

        :param name: [description]
        :type name: [type]
        :param config: [description]
        :type config: [type]
        :param description: [description], defaults to ''
        :type description: str, optional
        :return: [description]
        :rtype: [type]
        """
        return self._new_ref_config(
            ConfigRef.INPUT_PARAMS, name, config, description)

    def new_train_config(self, name, config, description=''):
        """
        Add a new JSON config for train_params

        :param name: [description]
        :type name: [type]
        :param config: [description]
        :type config: [type]
        :param description: [description], defaults to ''
        :type description: str, optional
        :return: [description]
        :rtype: [type]
        """
        return self._new_ref_config(
            ConfigRef.TRAIN_PARAMS, name, config, description)

    def new_model_params_config(self, name, config, description=''):
        """
        Add a new JSON config for model_params

        :param name: [description]
        :type name: [type]
        :param config: [description]
        :type config: [type]
        :param description: [description], defaults to ''
        :type description: str, optional
        :return: [description]
        :rtype: [type]
        """
        return self._new_ref_config(
            ConfigRef.MODEL_PARAMS, name, config, description)

    def json_data(self, val):
        return self.dbclient.df_to_json(val)

    def _new_ref_config(self, table, name, config, descripiton=''):
        new_obj = {
            RefAttr.NAME: name,
            RefAttr.CONFIG: config,
            RefAttr.DESC: descripiton
        }

        return self.dbclient.insert(
            table, new_obj).inserted_id

    def _new_config_from_components(self, dataset_params, input_params,
                                    architecture, model_params, train_params):
        return '{"{}":{}, "{}":{},"{}":{}, "{}":{}, "{}":{}}'.format(
            "dataset_params", dataset_params,
            "input_params", input_params,
            "architecture", architecture,
            "model_params", model_params,
            "train_params", train_params
        )



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"
__version__ = "0.0.1"


from .data_reader import *
from .data_generator import *
from .preprocessor import *



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"
__version__ = "0.0.1"


import pytest
from deoxys.keras.layers import Activation
from deoxys.model.activations import Activations
from deoxys.customize import register_activation, \
    unregister_activation, custom_activation
from deoxys.utils import Singleton


@pytest.fixture(autouse=True)
def clear_singleton():
    Singleton._instances = {}  # clear singleton


@pytest.fixture
def activation_class():
    class TestActivation(Activation):
        pass

    yield TestActivation


def test_is_singleton():
    activations_instance = Activations()
    another_instance = Activations()

    assert activations_instance is another_instance


def test_register_random_obj():
    with pytest.raises(ValueError):
        register_activation("TestActivation", object)


def test_register_activation_success(activation_class):
    register_activation("TestActivation", activation_class)

    assert Activations()._activations["TestActivation"] is activation_class


def test_register_duplicate_activation(activation_class):
    register_activation("TestActivation", activation_class)

    with pytest.raises(KeyError):
        register_activation("TestActivation", activation_class)


def test_unregister_activation(activation_class):
    register_activation("TestActivation", activation_class)

    assert Activations()._activations["TestActivation"] is activation_class
    unregister_activation("TestActivation")

    assert "TestActivation" not in Activations()._activations


def test_decorator():
    @custom_activation
    class TestActivation2(Activation):
        pass

    assert Activations()._activations["TestActivation2"] is TestActivation2



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"
__version__ = "0.0.1"


import pandas as pd


def read_csv(filepath, *args, **kwargs):
    # Load data to file
    df = pd.read_csv(filepath, *args, **kwargs)

    # Convert str-like array into float numbers
    object_keys = [key for key, dtype in df.dtypes.items()
                   if dtype == 'object']
    for key in object_keys:
        df[key] = df[key].map(_arraystr2float, na_action='ignore')

    return df


def _arraystr2float(val):
    if '"[' in val and ']"' in val:
        array = val[2: -2].split(',')
        if len(array) == 1:
            try:
                return float(array[0])
            except Exception:
                return array[0]
        else:
            return array
    else:
        return val



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"
__version__ = "0.0.1"


from ..model.optimizers import optimizer_from_config
from ..model.losses import loss_from_config
from ..model.metrics import metric_from_config
from ..utils import deep_copy


def load_params(model_params):
    """
    Create model objects from configurations

    :param model_params: dictionary of model objects (optimizer, loss, metrics)

    Example:
    ```
    {
        "loss": {
            "class_name": "BinaryFbetaLoss"
        },
        "optimizer": {
            "class_name": "adam",
            "config": {
                "learning_rate": 0.0001
            }
        },
        "metrics": [
            {
                "class_name": "BinaryFbeta"
            },
            "accuracy"
        ]
    }
    ```
    :type model_params: dict
    :return: [description]
    :rtype: dict
    """
    if not model_params:
        return {}
    params = deep_copy(model_params)
    for key, val in params.items():
        if key == 'optimizer':
            if type(val) == dict:
                params[key] = optimizer_from_config(val)
        elif key == 'loss':
            if type(val) == dict:
                params[key] = loss_from_config(val)
        elif key == 'metrics':
            for i, metric in enumerate(val):
                if type(metric) == dict:
                    params[key][i] = metric_from_config(metric)
                elif metric in ['TruePositives', 'FalsePositives',
                                'TrueNegatives', 'FalseNegatives']:
                    params[key][i] = metric_from_config(
                        {'class_name': metric, 'config': {'name': metric}}
                    )
    return params



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"
__version__ = "0.0.1"


from ..keras.utils import deserialize_keras_object
from ..keras.callbacks import *

import warnings
import numpy as np
import io
import csv
import os
import h5py
from collections import OrderedDict, Iterable

from ..utils import Singleton
from ..database import Tables, HDF5Attr, LogAttr


class DeoxysModelCallback(Callback):  # noqa: F405
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.deoxys_model = None

    def set_deoxys_model(self, deoxys_model):
        if not self.deoxys_model:
            self.deoxys_model = deoxys_model


class EvaluationCheckpoint(DeoxysModelCallback):
    """
    Evaluate test after some epochs. Only use when cross validation
    to avoid data leakage.
    """

    def __init__(self, filename=None, period=1,
                 separator=',', append=False):

        self.period = period
        self.epochs_since_last_save = 0

        self.sep = separator
        self.filename = filename
        self.append = append
        self.writer = None
        self.keys = None
        self.append_header = True

        self.file_flags = ''
        self._open_args = {'newline': '\n'}
        super().__init__()

    def on_train_begin(self, logs=None):
        if self.append:
            if os.path.exists(self.filename):
                with open(self.filename, 'r' + self.file_flags) as f:
                    self.append_header = not bool(len(f.readline()))
            mode = 'a'
        else:
            mode = 'w'
        self.csv_file = io.open(self.filename,
                                mode + self.file_flags,
                                **self._open_args)

    def on_epoch_end(self, epoch, logs=None):
        logs = logs or {}
        self.epochs_since_last_save += 1

        if self.epochs_since_last_save >= self.period:

            print('\nEvaluating test set...')
            self.epochs_since_last_save = 0
            score = self.deoxys_model.evaluate_test(verbose=1)

            def handle_value(k):
                is_zero_dim_ndarray = isinstance(k, np.ndarray) and k.ndim == 0
                if isinstance(k, str):
                    return k
                elif isinstance(k, Iterable) and not is_zero_dim_ndarray:
                    if k.ndim == 1:
                        return k[0]
                    else:
                        return '"[%s]"' % (', '.join(map(str, k)))
                else:
                    return k

            if self.keys is None:
                self.keys = [key for key in list(logs.keys())
                             if 'val_' not in key]

            if self.model.stop_training:
                # We set NA so that csv parsers do not fail in this last epoch.
                logs = dict([(k, logs[k] if k in logs else 'NA')
                             for k in self.keys])

            if not self.writer:
                class CustomDialect(csv.excel):
                    delimiter = self.sep
                fieldnames = ['epoch'] + self.keys

                self.writer = csv.DictWriter(self.csv_file,
                                             fieldnames=fieldnames,
                                             dialect=CustomDialect)
                if self.append_header:
                    self.writer.writeheader()

            row_dict = OrderedDict({'epoch': epoch})
            row_dict.update(
                (key, handle_value(score[i]))
                for i, key in enumerate(self.keys) if i < len(score))
            self.writer.writerow(row_dict)
            self.csv_file.flush()


class DBLogger(Callback):  # noqa: F405

    def __init__(self, dbclient, session):
        """
        Log performance to database

        :param dbclient: the database client that stores all data
        :type dbclient: deoxys.database.DBClient
        :param session: session id
        :type session: str, int, or ObjectId, depending of the provider
                       of DBClient
        """
        self.dbclient = dbclient
        self.session = session

        self.keys = None

        super().__init__()

    def on_epoch_end(self, epoch, logs=None):
        logs = logs or {}

        def handle_value(k):
            is_zero_dim_ndarray = isinstance(k, np.ndarray) and k.ndim == 0
            if isinstance(k, str):
                return k
            elif isinstance(k, Iterable) and not is_zero_dim_ndarray:
                if k.ndim == 1:
                    if isinstance(k[0], np.generic):
                        return np.asscalar(k[0])
                    else:
                        return k[0]
                else:
                    return '"[%s]"' % (', '.join(map(str, k)))
            else:
                if isinstance(k, np.generic):
                    return np.asscalar(k)
                else:
                    return k

        if self.keys is None:
            self.keys = sorted(logs.keys())

        if self.model.stop_training:
            # We set NA so that it won't fail in this last epoch.
            logs = dict([(k, logs[k] if k in logs else 'NA')
                         for k in self.keys])

        identifier = {LogAttr.SESSION_ID: self.session,
                      LogAttr.EPOCH: epoch + 1}
        perf_log = OrderedDict(identifier)
        perf_log.update((key, handle_value(logs[key])) for key in self.keys)

        self.dbclient.update_insert(Tables.LOGS, identifier, perf_log)


class PredictionCheckpoint(DeoxysModelCallback):
    """
    Predict test in every number of epochs
    """

    def __init__(self, filepath=None, period=1, use_original=False,
                 dbclient=None, session=None):
        self.period = period
        self.epochs_since_last_save = 0

        self.filepath = filepath
        self.use_original = use_original

        self.dbclient = dbclient
        self.session = session

        super().__init__()

    def on_epoch_end(self, epoch, logs=None):
        logs = logs or {}
        self.epochs_since_last_save += 1

        if self.epochs_since_last_save >= self.period:
            self.epochs_since_last_save = 0

            print('\nPredicting validation data...')

            # Predict all data
            predicted = self.deoxys_model.predict_val(verbose=1)

            # Get file name
            filepath = self.filepath.format(epoch=epoch + 1, **logs)

            # Create the h5 file
            hf = h5py.File(filepath, 'w')
            hf.create_dataset('predicted', data=predicted, compression="gzip")
            hf.close()

            if self.use_original:
                original_data = self.deoxys_model.data_reader.original_val

                for key, val in original_data.items():
                    hf = h5py.File(filepath, 'a')
                    hf.create_dataset(key, data=val, compression="gzip")
                    hf.close()
            else:
                # Create data from val_generator
                x = None
                y = None

                val_gen = self.deoxys_model.data_reader.val_generator
                data_gen = val_gen.generate()

                for _ in range(val_gen.total_batch):
                    next_x, next_y = next(data_gen)
                    if x is None:
                        x = next_x
                        y = next_y
                    else:
                        x = np.concatenate((x, next_x))
                        y = np.concatenate((y, next_y))

                hf = h5py.File(filepath, 'a')
                hf.create_dataset('x', data=x, compression="gzip")
                hf.create_dataset('y', data=y, compression="gzip")
                hf.close()

            if self.dbclient:
                item = OrderedDict(
                    {HDF5Attr.SESSION_ID: self.session,
                     HDF5Attr.EPOCH: epoch + 1})
                item.update(
                    {HDF5Attr.FILE_LOCATION: os.path.abspath(filepath)})

                self.dbclient.insert(Tables.PREDICTIONS, item)


class DeoxysModelCheckpoint(DeoxysModelCallback,
                            ModelCheckpoint):  # noqa: F405

    def __init__(self, filepath, monitor='val_loss', verbose=0,
                 save_best_only=False, save_weights_only=False,
                 mode='auto', period=1,
                 dbclient=None, session=None):
        super().__init__(filepath=filepath,
                         monitor=monitor, verbose=verbose,
                         save_best_only=save_best_only,
                         save_weights_only=save_weights_only,
                         mode=mode, period=period)

        self.dbclient = dbclient
        self.session = session

    def on_epoch_end(self, epoch, logs=None):
        logs = logs or {}
        self.epochs_since_last_save += 1
        if self.epochs_since_last_save >= self.period:
            self.epochs_since_last_save = 0
            filepath = self.filepath.format(epoch=epoch + 1, **logs)
            abs_path = os.path.abspath(filepath)
            if self.save_best_only:
                current = logs.get(self.monitor)
                if current is None:
                    warnings.warn('Can save best model '
                                  ' only with % s available, '
                                  'skipping.' % (self.monitor), RuntimeWarning)
                else:
                    if self.monitor_op(current, self.best):
                        if self.verbose > 0:
                            print('\nEpoch %05d: %s improved from '
                                  '%0.5f to %0.5f,'
                                  ' saving model to %s'
                                  % (epoch + 1, self.monitor, self.best,
                                     current, filepath))
                        self.best = current
                        if self.save_weights_only:
                            self.model.save_weights(filepath, overwrite=True)
                        else:
                            self.deoxys_model.save(filepath, overwrite=True)

                        if self.dbclient:
                            item = OrderedDict(
                                {HDF5Attr.SESSION_ID: self.session,
                                 HDF5Attr.EPOCH: epoch + 1})
                            item.update(
                                {HDF5Attr.FILE_LOCATION: abs_path})

                            self.dbclient.insert(Tables.MODELS, item)
                    else:
                        if self.verbose > 0:
                            print('\nEpoch %05d: %s did not improve from '
                                  '%0.5f' %
                                  (epoch + 1, self.monitor, self.best))
            else:
                if self.verbose > 0:
                    print('\nEpoch %05d: saving model to %s' %
                          (epoch + 1, filepath))
                if self.save_weights_only:
                    self.model.save_weights(filepath, overwrite=True)
                else:
                    self.deoxys_model.save(filepath, overwrite=True)

                if self.dbclient:
                    item = OrderedDict(
                        {HDF5Attr.SESSION_ID: self.session,
                         HDF5Attr.EPOCH: epoch + 1})
                    item.update({HDF5Attr.FILE_LOCATION: abs_path})
                    self.dbclient.insert(Tables.MODELS, item)


class Callbacks(metaclass=Singleton):
    """
    A singleton that contains all the registered customized callbacks
    """

    def __init__(self):
        self._callbacks = {
        }

    def register(self, key, callback):
        if not issubclass(callback, Callback):  # noqa: F405
            raise ValueError(
                "The customized callback has to be a subclass"
                + " of keras.callbacks.Callback"
            )

        if key in self._callbacks:
            raise KeyError(
                "Duplicated key, please use another key for this callback"
            )
        else:
            self._callbacks[key] = callback

    def unregister(self, key):
        if key in self._callbacks:
            del self._callbacks[key]

    @property
    def callbacks(self):
        return self._callbacks


def register_callback(key, callback):
    """
    Register the customized callback.
    If the key name is already registered, it will raise a KeyError exception

    :param key: the unique key-name of the callback
    :type key: str
    :param callback: the customized callback class
    :type callback: keras.callbacks.Callback
    """
    Callbacks().register(key, callback)


def unregister_callback(key):
    """
    Remove the registered callback with the key-name

    :param key: the key-name of the callback to be removed
    :type key: str
    """
    Callbacks().unregister(key)


def callback_from_config(config):
    if 'class_name' not in config:
        raise ValueError('class_name is needed to define callback')

    if 'config' not in config:
        # auto add empty config for callback with only class_name
        config['config'] = {}
    return deserialize_keras_object(config,
                                    module_objects=globals(),
                                    custom_objects=Callbacks().callbacks,
                                    printable_module_name='callback')



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"
__version__ = "0.0.1"


import tensorflow as tf

import json
import h5py
import numpy as np
from itertools import product

from ..keras.models import \
    model_from_config as keras_model_from_config, \
    model_from_json as keras_model_from_json, \
    load_model as keras_load_model, Model as KerasModel, \
    clone_model
from ..utils import is_default_tf_eager_mode

from ..keras import backend as K

from ..loaders import load_architecture, load_params, \
    load_data, load_train_params
from ..utils import load_json_config
from .layers import Layers
from .metrics import Metrics
from .losses import Losses
from .activations import Activations
from .callbacks import DeoxysModelCallback


class Model:
    """
    Model
    """
    _evaluate_param_keys = ['callbacks', 'max_queue_size',
                            'workers', 'use_multiprocessing', 'verbose']

    _predict_param_keys = ['callbacks', 'max_queue_size',
                           'workers', 'use_multiprocessing', 'verbose']

    _fit_param_keys = ['epochs', 'verbose', 'callbacks', 'class_weight',
                       'max_queue_size', 'workers',
                       'use_multiprocessing', 'shuffle', 'initial_epoch']

    def __init__(self, model, model_params=None, train_params=None,
                 data_reader=None,
                 pre_compiled=False, weights_file=None, config=None):
        """
        Create a deoxys model

        :param model: a keras model object
        :type model: tf.keras.models.Model, keras.models.Model
        :param model_params: params to compile a keras model, defaults to None
        :type model_params: dict, optional
        :param train_params: params for training, evaluate, predict,
        defaults to None
        :type train_params: dict, optional
        :param data_reader: A deoxys data reader, defaults to None
        :type data_reader: deoxys.data.DataReader, optional
        :param pre_compiled: True if model has been compiled, defaults to False
        :type pre_compiled: bool, optional
        :param weights_file: path to h5 file that contains the weights of the
        keras model, defaults to None
        :type weights_file: str, optional
        :param config: full config to create the model, defaults to None
        :type config: dict, optional
        :raises ValueError: raises error if model_params is defined without
        optimizer
        """

        self._model = model
        self._model_params = model_params
        self._train_params = train_params
        self._compiled = pre_compiled
        self._data_reader = data_reader
        self._layers = None

        self.config = config or {}

        if model_params:
            if 'optimizer' in model_params:
                self._model.compile(**model_params)

                if weights_file:
                    # This initializes the variables used by the optimizers,
                    # as well as any stateful metric variables
                    self._model.train_on_batch(
                        *next(self.data_reader.train_generator.generate()))

                    self._model.load_weights(weights_file)
                self._compiled = True
            else:
                raise ValueError('optimizer is a required parameter in '
                                 'model_params.')

    def compile(self, optimizer=None, loss=None, metrics=None,
                loss_weights=None,
                sample_weight_mode=None, weighted_metrics=None,
                target_tensors=None, **kwargs):
        """Compile the model

        :raises Warning: calling this function will recompile the model with
        new configuration
        """
        if self._compiled:
            raise Warning(
                'This will override the previous configuration of the model.')

        self._model.compile(optimizer, loss, metrics,
                            loss_weights,
                            sample_weight_mode, weighted_metrics,
                            target_tensors, **kwargs)
        self._compiled = True

    def save(self, filename, *args, **kwargs):
        """
        Save model to file

        :param filename: name of the file
        :type filename: str
        """
        self._model.save(filename, *args, **kwargs)

        if self.config:
            config = json.dumps(self.config)

            saved_model = h5py.File(filename, 'a')
            saved_model.attrs.create('deoxys_config', config)
            saved_model.close()

    def fit(self, *args, **kwargs):
        """
        Train model
        """
        # Reset layer map as weight will change after training
        self._layers = None

        self._model.fit(*args, **kwargs)

    def predict(self, *args, **kwargs):
        """
        Predict model
        """
        return self._model.predict(*args, **kwargs)

    def evaluate(self, *args, **kwargs):
        """
        Evaluate model
        """
        return self._model.evaluate(*args, **kwargs)

    def fit_generator(self, *args, **kwargs):
        # Reset layer map as weight will change after training
        self._layers = None
        return self._model.fit_generator(*args, **kwargs)

    def evaluate_generator(self, *args, **kwargs):
        return self._model.evaluate_generator(*args, **kwargs)

    def predict_generator(self, *args, **kwargs):
        return self._model.predict_generator(*args, **kwargs)

    def fit_train(self, **kwargs):
        """
        Train the model with training data
        """
        # Reset layer map as weight will change after training
        self._layers = None

        params = self._get_train_params(self._fit_param_keys, **kwargs)
        train_data_gen = self._data_reader.train_generator
        train_steps_per_epoch = train_data_gen.total_batch

        val_data_gen = self._data_reader.val_generator
        val_steps_per_epoch = val_data_gen.total_batch

        return self.fit_generator(train_data_gen.generate(),
                                  steps_per_epoch=train_steps_per_epoch,
                                  validation_data=val_data_gen.generate(),
                                  validation_steps=val_steps_per_epoch,
                                  **params)

    def evaluate_train(self, **kwargs):
        """
        Evaluate the model on the training data
        """
        params = self._get_train_params(self._evaluate_param_keys, **kwargs)
        data_gen = self._data_reader.train_generator
        steps_per_epoch = data_gen.total_batch

        return self.evaluate_generator(data_gen.generate(),
                                       steps=steps_per_epoch,
                                       **params)

    def evaluate_val(self, **kwargs):
        """
        Evaluate model's performance using validation data
        """
        params = self._get_train_params(self._evaluate_param_keys, **kwargs)
        data_gen = self._data_reader.val_generator
        steps_per_epoch = data_gen.total_batch

        return self.evaluate_generator(data_gen.generate(),
                                       steps=steps_per_epoch,
                                       **params)

    def predict_val(self, **kwargs):
        """
        Predict validation data
        """
        params = self._get_train_params(self._predict_param_keys, **kwargs)
        data_gen = self._data_reader.val_generator
        steps_per_epoch = data_gen.total_batch

        return self.predict_generator(data_gen.generate(),
                                      steps=steps_per_epoch,
                                      **params)

    def evaluate_test(self, **kwargs):
        """
        Evaluate model performance using test data
        """
        params = self._get_train_params(self._evaluate_param_keys, **kwargs)
        data_gen = self._data_reader.test_generator
        steps_per_epoch = data_gen.total_batch

        return self.evaluate_generator(data_gen.generate(),
                                       steps=steps_per_epoch,
                                       **params)

    def predict_test(self, **kwargs):
        """
        Predict test data
        """
        params = self._get_train_params(self._predict_param_keys, **kwargs)
        data_gen = self._data_reader.test_generator
        steps_per_epoch = data_gen.total_batch

        return self.predict_generator(data_gen.generate(),
                                      steps=steps_per_epoch,
                                      **params)

    @property
    def is_compiled(self):
        return self._compiled

    @property
    def model(self):
        """
        Return the keras model
        """
        return self._model

    @property
    def data_reader(self):
        """
        Get the data reader used in this model
        """
        return self._data_reader

    @property
    def layers(self):
        """
        Get the dictionary of layers in the model

        :return: dictionary of layers
        :rtype: dict
        """
        if self._layers is None:
            self._layers = {layer.name: layer for layer in self.model.layers}

        return self._layers

    @property
    def node_graph(self):
        """
        Node graph from nodes in model, ignoring resize and concatenate nodes
        """
        layers = self.layers

        connection = []

        def previous_layers(name):
            # Keep looking for the previous layer of resize layers
            if 'resize' in name:
                prevs = layers[name].inbound_nodes[0].get_config()[
                    'inbound_layers']
                if type(prevs) == str:
                    prev = prevs
                else:
                    # in case there are multiple layers, take the 1st one
                    prev = prevs[0]
                return previous_layers(prev)
            return name

        model = self.model

        for layer in model.layers:
            if 'resize' in layer.name or 'concatenate' in layer.name:
                continue
            inbound_nodes = layer.inbound_nodes
            for node in inbound_nodes:
                inbound_layers = node.get_config()['inbound_layers']

                if type(inbound_layers) == str:
                    if 'concatenate' in inbound_layers:
                        concat_layer = layers[inbound_layers]

                        nodes = concat_layer.inbound_nodes[0].get_config()[
                            'inbound_layers']
                        for n in nodes:
                            options = {}
                            prev_layer = previous_layers(n)
                            if 'transpose' not in prev_layer:
                                options.update({'length': 300})
                            connection.append({
                                'from': prev_layer,
                                'to': layer.name,
                                **options
                            })
                    else:
                        connection.append({
                            'from': layers[inbound_layers].name,
                            'to': layer.name,
                        })
        return connection

    def sub_model(self, layer_name):
        """
        Create a sub-model with the same inputs, and the outputs of a specific
        layer in the deoxys model.

        :param layer_name: name of layer
        :type layer_name: str
        :return: Model, whose outputs are of the layer_name
        :rtype: tf.keras.models.Model or keras.models.Model
        """
        return KerasModel(inputs=self.model.inputs,
                          outputs=self.layers[layer_name].output)

    def activation_map(self, layer_name, images):
        """
        Get activation map of a list of images
        """
        model = self.sub_model(layer_name)
        res = model.predict(images, verbose=1)

        del model
        return res

    def _get_gradient_loss(self, outputs, filter_index, loss_fn):
        if loss_fn is None:
            loss_value = tf.reduce_mean(
                outputs[..., filter_index])
        else:
            loss_value = loss_fn(outputs)

        return loss_value

    def activation_maximization(self, layer_name, img=None,
                                step_size=1, epochs=20,
                                filter_index=0, loss_fn=None,
                                verbose=True):
        """
        Return the image that maximize the activation output of one or more
        filters in a specific layer.

        :param layer_name: name of the node
        :type layer_name: str
        :param img: list of initial images, defaults to None. If None, a random
        image with noises will be used
        :type img: [list], optional
        :param step_size: Size of the step when performing gradient descent,
        defaults to 1
        :type step_size: int, optional
        :param epochs: Number of epochs for gradient descent, defaults to 20
        :type epochs: int, optional
        :param filter_index: index of the filter to get the gradient, can be
        any number between 0 and (size of the filters - 1), defaults to 0
        :type filter_index: int, or list, optional
        :return: list of images that maximize the activation's filters
        :rtype: list
        """
        if type(filter_index) == int:
            list_index = [filter_index]
        else:
            list_index = filter_index

        input_shape = [1] + (self.model.input.shape)[1:]
        if img is None:
            input_img_data = np.random.random(input_shape)
        else:
            input_img_data = img

        input_img_data = [tf.Variable(
            tf.cast(input_img_data, K.floatx())) for _ in list_index]

        activation_model = self.sub_model(layer_name)

        for _ in range(epochs):
            if verbose:
                print('epoch', _, '/', epochs)

            for i, filter_index in enumerate(list_index):
                if verbose:
                    print('filter', filter_index)
                if is_default_tf_eager_mode():

                    with tf.GradientTape() as tape:
                        outputs = activation_model(input_img_data[i])

                        loss_value = self._get_gradient_loss(
                            outputs, filter_index, loss_fn)

                    grads = tape.gradient(loss_value, input_img_data[i])
                else:
                    outputs = activation_model.output

                    loss_value = self._get_gradient_loss(
                        outputs, filter_index, loss_fn)

                    gradient = K.gradients(loss_value, activation_model.input)

                    grads = K.function(activation_model.input,
                                       gradient)(input_img_data)[0]

                normalized_grads = grads / \
                    (tf.sqrt(tf.reduce_mean(tf.square(grads))) + 1e-5)

                input_img_data[i].assign_add(normalized_grads * step_size)

        if len(input_img_data) > 1:
            if is_default_tf_eager_mode():
                return [K.get_value(input_img) for input_img in input_img_data]
            else:
                return [K.eval(input_img) for input_img in input_img_data]

        if is_default_tf_eager_mode():
            return K.get_value(input_img_data[0])
        else:
            return K.eval(input_img_data[0])

    def _get_backprop_loss(self, output, mode='max', output_index=0,
                           loss_fn=None):
        if mode == 'max':
            loss = K.max(output, axis=-1)
        elif mode == 'mean':
            loss = K.mean(output, axis=-1)
        elif mode == 'min':
            loss = K.min(output, axis=-1)
        elif mode == 'one':
            loss = output[..., output_index]
        elif mode == 'all':
            loss = output
        elif loss_fn is not None:
            loss = loss_fn(output)
        else:
            loss = output

        return loss

    def _max_filter_map(self, output):
        return K.argmax(output, axis=-1)

    def _backprop_eagerly(self, layer_name, images, mode='max',
                          output_index=0, loss_fn=None):
        img_tensor = tf.Variable(tf.cast(images, K.floatx()))
        activation_model = self.sub_model(layer_name)
        with tf.GradientTape() as tape:
            tape.watch(img_tensor)
            output = activation_model(img_tensor)

            loss = self._get_backprop_loss(output, mode, output_index, loss_fn)

        grads = tape.gradient(loss, img_tensor)

        return K.get_value(grads)

    def _backprop_symbolic(self, layer_name, images, mode='max',
                           output_index=0, loss_fn=None):
        output = self.layers[layer_name].output

        loss = self._get_backprop_loss(output, mode, output_index, loss_fn)

        grads = K.gradients(loss, self.model.input)[0]

        fn = K.function(self.model.input, grads)

        return fn(images)

    def backprop(self, layer_name, images, mode='max', output_index=0,
                 loss_fn=None):
        """
        Return saliency map, or backprop, or gradient map of a list of images

        :param layer_name: name of the layer
        :type layer_name: str
        :param images: list of images
        :type images: list
        :param mode: mode to calculate the loss to use when backpropagation,
        defaults to 'max', other options are 'mean', 'min', 'one', 'custom',
        and 'all'.

        'max', 'min', 'mean': calculate the loss by calculating the max, min,
        or mean over the inner most axis (axis=-1) of the output.

        'one': calculate the loss on one index in the inner most axis (axis=-1)
        of the output of the layer.

        'all': use the output of the layer as the loss score.

        'custom': use a custom function to calculate the loss score based on
        the output of the layer.

        :type mode: str, optional
        :param output_index: use when mode='one', defaults to 0
        :type output_index: int, optional
        :param loss_fn: use when mode='custom', the function to calculate the
        loss score based on the output of the layer, defaults to None
        :type loss_fn: function, optional
        :return: resulting images when performing backpropagation.
        :rtype: numpy.array of images
        """
        if is_default_tf_eager_mode():
            grads = self._backprop_eagerly(
                layer_name, images, mode, output_index, loss_fn)
        else:
            grads = self._backprop_symbolic(
                layer_name, images, mode, output_index, loss_fn)
        return grads

    def _gradient_backprop(self, gradient_name, layer_name,
                           images, mode, output_index, loss_fn=None):
        # save current weight
        weights = self.model.get_weights()

        with tf.Graph().as_default() as g:
            with g.gradient_override_map({'Relu': gradient_name}):
                tf.compat.v1.experimental.output_all_intermediates(True)
                new_model = clone_model(self.model)
                # Apply weights
                new_model.set_weights(weights)

                output = self._get_backprop_loss(
                    new_model.get_layer(layer_name).output,
                    mode, output_index, loss_fn)

                # if mode == 'max':
                #     output = K.max(new_model.get_layer(
                #         layer_name).output, axis=-1)
                # elif mode == 'one':
                #     output = new_model.get_layer(
                #         layer_name).output[..., output_index]
                # elif mode == 'all':
                #     output = new_model.get_layer(
                #         layer_name).output
                # elif loss_fn is not None:
                #     output = loss_fn(new_model.get_layer(layer_name).output)
                # else:
                #     output = new_model.get_layer(
                #         layer_name).output

                grads = K.gradients(output, new_model.input)[0]

                fn = K.function(new_model.input, grads)

                grad_output = fn(images)

            del new_model
        del g

        return grad_output

    def deconv(self, layer_name, images, mode='max',
               output_index=0, loss_fn=None):
        if is_default_tf_eager_mode():
            return self._gradient_backprop_eager(
                _DeconvRelu, layer_name,
                images, mode, output_index, loss_fn
            )
        else:
            return self._gradient_backprop('DeconvNet', layer_name,
                                           images, mode, output_index, loss_fn)

    def guided_backprop(self, layer_name, images, mode='max',
                        output_index=0, loss_fn=None):
        if is_default_tf_eager_mode():
            return self._gradient_backprop_eager(
                _GuidedBackPropRelu, layer_name,
                images, mode, output_index, loss_fn
            )
        else:
            return self._gradient_backprop('GuidedBackProp', layer_name,
                                           images, mode, output_index, loss_fn)

    def _gradient_backprop_eager(self, grad_fn, layer_name, images, mode='max',
                                 output_index=0, loss_fn=None):
        # save current weight
        weights = self.model.get_weights()

        new_model = clone_model(self.model)
        # Apply weights
        new_model.set_weights(weights)

        for layer in new_model.layers:
            if 'activation' in layer.get_config():
                if 'relu' in layer.activation.__name__:
                    layer.activation = grad_fn

        guided_model = KerasModel(new_model.inputs,
                                  new_model.get_layer(layer_name).output)

        img_tensor = tf.Variable(tf.cast(images, K.floatx()))
        with tf.GradientTape() as tape:
            tape.watch(img_tensor)
            output = guided_model(img_tensor)

            loss = self._get_backprop_loss(output, mode, output_index, loss_fn)

        grads = tape.gradient(loss, img_tensor)

        del guided_model
        del new_model

        return K.get_value(grads)

    def max_filter(self, layer_name, images):
        """
        Return a list of images in which each pixel value is the index of the
        filter having the max value in the activation map.
        """
        return self._max_filter_map(
            self.activation_map(layer_name, images))

    def _get_train_params(self, keys, **kwargs):
        params = {}

        # Load train_params every run
        train_params = load_train_params(self._train_params)

        if 'callbacks' in train_params and 'callbacks' in kwargs:
            kwargs['callbacks'] = train_params['callbacks'] + \
                kwargs['callbacks'] if type(
                kwargs['callbacks']) == list else [kwargs['callbacks']]

        params.update(train_params)
        params.update(kwargs)

        if 'callbacks' in params:
            # set deoxys model for custom model
            for callback in params['callbacks']:
                if isinstance(callback, DeoxysModelCallback):
                    callback.set_deoxys_model(self)

        params = {key: params[key] for key in params if key in keys}
        return params


def model_from_full_config(model_config, weights_file=None, **kwargs):
    """
    Return the model from the full config

    :param model_config: a json string or a dictionary contains the
    architecture, model_params, input_params of the model
    :type model_config: a JSON string or a dictionary object
    :return: The model
    :rtype: deoxys.model.Model
    """
    config = load_json_config(model_config)

    if ('architecture' not in config or 'input_params' not in config):
        raise ValueError('architecture and input_params are required')

    return model_from_config(
        config['architecture'],
        config['input_params'],
        config['model_params'] if 'model_params' in config else None,
        config['train_params'] if 'train_params' in config else None,
        config['dataset_params'] if 'dataset_params' in config else None,
        weights_file=weights_file,
        **kwargs)


def model_from_config(architecture, input_params,
                      model_params=None, train_params=None,
                      dataset_params=None, weights_file=None, **kwargs):
    architecture, input_params, model_params, train_params = load_json_config(
        architecture, input_params, model_params, train_params)

    config = {
        'architecture': architecture,
        'input_params': input_params,
        'model_params': model_params,
        'train_params': train_params,
        'dataset_params': dataset_params
    }

    # load the model based on the architecture type (Unet / Dense/ Sequential)
    loaded_model = load_architecture(architecture, input_params)

    # Load the parameters to compile the model
    loaded_params = load_params(model_params)

    # the keyword arguments will replace existing params
    loaded_params.update(kwargs)

    # load the data generator
    data_generator = None
    if dataset_params:
        data_generator = load_data(dataset_params)

    return Model(loaded_model, loaded_params, train_params,
                 data_generator, config=config, weights_file=weights_file,
                 **kwargs)


def model_from_keras_config(config, **kwarg):
    return Model(keras_model_from_config(config), **kwarg)


def model_from_keras_json(json, **kwarg):
    return Model(keras_model_from_json(json), **kwarg)


def load_model(filename, **kwargs):
    """
    Load model from file

    :param filename: path to the h5 file
    :type filename: str
    """
    # Keras got the error of loading custom object
    try:
        model = Model(keras_load_model(filename,
                                       custom_objects={
                                           **Layers().layers,
                                           **Activations().activations,
                                           **Losses().losses,
                                           **Metrics().metrics}),
                      pre_compiled=True, **kwargs)

        with h5py.File(filename) as hf:
            if 'deoxys_config' in hf.attrs.keys():
                config = hf.attrs['deoxys_config']
                model._data_reader = load_data(config['dataset_params'])

    except Exception:
        hf = h5py.File(filename, 'r')
        if 'deoxys_config' in hf.attrs.keys():
            config = hf.attrs['deoxys_config']
        hf.close()

        model = model_from_full_config(config, weights_file=filename)

    return model


@tf.RegisterGradient("GuidedBackProp")
def _GuidedBackProp(op, grad):
    dtype = op.inputs[0].dtype
    return grad * tf.cast(grad > 0., dtype) * tf.cast(op.inputs[0] > 0., dtype)


@tf.RegisterGradient("DeconvNet")
def _DeconvNet(op, grad):
    dtype = op.inputs[0].dtype
    return grad * tf.cast(grad > 0., dtype)


@tf.RegisterGradient("BackProp")
def _BackProp(op, grad):
    dtype = op.inputs[0].dtype
    return grad * tf.cast(op.inputs[0] > 0., dtype)


@tf.custom_gradient
def _GuidedBackPropRelu(x):
    res = K.relu(x)

    def grad(dy):
        return dy * tf.cast(dy > 0., x.dtype) * tf.cast(x > 0., x.dtype)

    return res, grad


@tf.custom_gradient
def _DeconvRelu(x):
    res = K.relu(x)

    def grad(dy):
        return dy * tf.cast(dy > 0., x.dtype)

    return res, grad


@tf.custom_gradient
def _BackPropRelu(x):
    res = K.relu(x)

    def grad(dy):
        return dy * tf.cast(x > 0., x.dtype)

    return res, grad



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"
__version__ = "0.0.1"


import pytest
from deoxys.keras.callbacks import CSVLogger, ModelCheckpoint, \
    TerminateOnNaN
from deoxys.loaders import load_train_params
from deoxys.utils import read_file, load_json_config


def test_load_params():
    params = load_train_params(load_json_config(
        read_file('tests/json/train_param_config.json')))

    assert params['epochs'] == 5
    assert len(params['callbacks']) == 3

    assert isinstance(params['callbacks'][0], CSVLogger)
    assert isinstance(params['callbacks'][1], ModelCheckpoint)
    assert isinstance(params['callbacks'][2], TerminateOnNaN)



def read_file(filename):
    """
    Return the content of a file

    :param filename: file name
    :type filename: str
    :return: content of the file
    :rtype: str
    """
    with open(filename, "r", encoding="utf-8") as f:
        return f.read()


def write_file(content, filename):
    """
    Write to a file

    :param content: content of the file
    :type content: str
    :param filename: file name
    :type filename: str
    """
    with open(filename, "w", encoding="utf-8") as f:
        f.write(content)


def write_byte(content, filename):
    """
    Write bytes to a file

    :param content: content of the file
    :type content: bytes
    :param filename: file name
    :type filename: str
    """
    with open(filename, "wb") as f:
        f.write(content)



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"
__version__ = "0.0.1"


from ..data.data_reader import datareader_from_config
from ..data.preprocessor import preprocessor_from_config
from ..utils import deep_copy


def load_data(dataset_params):
    """
    Create an data reader instance from a config object.
    Example:
    {
        "class_name": "HDF5Reader",
        "config": {
            "filename": "../../dataset.h5",
            "batch_size": 16,
            "x_name": "input",
            "y_name": "target",
            "batch_cache": 4,
            "train_folds": [0, 1],
            "val_folds": [2],
            "test_folds": [3, 4],
            "preprocessors": {
                "class_name": "WindowingPreprocessor",
                "config": {
                    "window_center": 1094,
                    "window_width": 200,
                    "channel": 0
                }
            }
        }
    }

    :param dataset_params: the configuration of the data reader
    :type dataset_params: dict
    :return: the data reader instance
    :rtype: deoxys.data.DataReader
    """
    if not dataset_params:
        return {}
    params = deep_copy(dataset_params)

    if params['config']:
        for key, val in params['config'].items():
            if key == 'preprocessors':
                if type(val) == dict:
                    params['config'][key] = [preprocessor_from_config(val)]
                if type(val) == list:
                    preprocessors = []
                    for v in val:
                        if type(v) == dict:
                            preprocessors.append(preprocessor_from_config(v))
                        elif type(v) == str:
                            preprocessors.append(
                                preprocessor_from_config({'class_name': v}))
                    params['config'][key] = preprocessors

    return datareader_from_config(params)



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"
__version__ = "0.0.1"


from sanic import Sanic
from sanic.response import file, json, html
import pkgutil
import os
import matplotlib.pyplot as plt
import mpld3
from ..experiment import MultiExperimentDB
from ..utils import write_byte, read_file
from ..model import load_model


class BaseApp:
    def __init__(self):
        self.app = Sanic()

        # @self.app.route("/")
        # async def test(request):
        #     return json({"hello": "world"})
        self.working_model = load_model(
            # '../../hn_perf/logs_db/'
            # '5e3547f57a4c104db6d4c5d2/model/model.004.h5'
            '../../oxford_perf/logs_db/'
            '5e2e2349a356a4893813c8f7/model/model.022.h5'
        )

    def add_route_handler(self, uri, handler, *args, **kwargs):
        self.app.route(uri, *args, **kwargs)(handler)

    def run(self):
        self.app.run(host="0.0.0.0", port=8000)


class VisApp(BaseApp):
    def __init__(self, dbclient, basepath='.', provider='mongodb'):
        super().__init__()
        self.me = MultiExperimentDB(dbclient)
        self.basepath = basepath

        if 'env' in os.environ and os.environ['ENV'] == 'dev':
            self._initialize_dev_mode()
        else:
            self._generate_static_assets()
            self._map_static_files()

            self.create_home_page()
        self.add_routes()

    def create_home_page(self):
        html_content = read_file(os.path.join(
            self.basepath, 'assets/html/index.html'))

        @self.app.route('/')
        async def handler(resquest):
            return html(html_content)

    def add_routes(self):
        @self.app.route('/api/experiments')
        async def all_experiment(request):
            exps = self.me.experiments

            data = self.me.json_data(exps)

            return json(data)

        @self.app.route('/api/sessions')
        async def all_sessions(request):
            headers = request.headers
            exp_id = headers['experiment_id']

            sessions = self.me.sessions_from_experiments(exp_id)

            data = self.me.json_data(sessions)

            return json(data)

        @self.app.route('/api/graph_nodes')
        def graph_nodes(request):
            # headers = request.headers
            # exp_id = headers['experiment_id']
            graph_nodes = self.working_model.node_graph
            layers = self.working_model.layers
            nodes = [{'id': layer,
                      'label': layer + '\n' + str(layers[layer].output.shape)}
                     for layer in layers
                     if 'resize' not in layer and 'concat' not in layer]

            return json({
                'nodes': nodes,
                'edges': graph_nodes
            })

        @self.app.route('/api/activation_map')
        def activation_map(request):
            headers = request.headers
            layer = headers['layer_name']
            img_id = int(headers['img_id'])

            model = self.working_model
            dr = model.data_reader

            batch_size = dr.batch_size
            batch_index = img_id // batch_size
            batch_pos = img_id % batch_size
            print(batch_index, batch_pos)
            for i, img in enumerate(dr.val_generator.generate()):
                if i == batch_index:
                    image = img[0][batch_pos:batch_pos + 1]
                    break

            print(image.shape)
            print('getting activation map')
            activation_map = model.activation_map(layer, image)[0]

            print('plotting')
            filter_num = activation_map.shape[-1]
            filters = []
            img_size = activation_map.shape[:-1]
            figsize = (img_size[0] / 50, img_size[1] / 50)
            if figsize[0] > 3 or figsize[1] > 3:
                figsize = (3, 3 * img_size[1] / img_size[0])
            elif figsize[0] < 2 or figsize[1] < 2:
                figsize = (1, 1 * img_size[1] / img_size[0])
            for i in range(filter_num):
                fig = plt.figure(figsize=figsize)
                # if image.shape[-1] == 3:
                #     plt.imshow(image[0])
                # else:
                #     plt.imshow(image[0][..., 0])
                plt.imshow(activation_map[..., i], alpha=0.3, cmap='gray')
                mpld3.plugins.connect(fig, mpld3.plugins.BoxZoom(False, False))
                filters.append(mpld3.fig_to_html(fig))
                plt.close('all')

            return json(filters)

    def _initialize_dev_mode(self):
        @self.app.route('/')
        async def handler(resquest):
            html_file = pkgutil.get_data(__name__, "templates/html/index.html")

            return html(str(html_file, 'utf-8'))

        self.app.static('/static/js',
                        os.path.join(os.path.dirname(__file__),
                                     "templates/js"), name='static/js')
        self.app.static('/static/css',
                        os.path.join(os.path.dirname(__file__),
                                     "templates/css"), name='static/css')
        self.app.static('/static/html',
                        os.path.join(os.path.dirname(__file__),
                                     "templates/html"), name='static')
        self.app.url_for('static', filename='any')

    def _generate_static_assets(self):
        basepath = self.basepath

        html_path = os.path.join(basepath, 'assets/html')
        js_path = os.path.join(basepath, 'assets/js')
        css_path = os.path.join(basepath, 'assets/css')

        if not os.path.exists(basepath):
            os.makedirs(basepath)

        if not os.path.exists(html_path):
            os.makedirs(html_path)

        if not os.path.exists(js_path):
            os.makedirs(js_path)

        if not os.path.exists(css_path):
            os.makedirs(css_path)

        html_file = pkgutil.get_data(__name__, "templates/html/index.html")
        write_byte(html_file, os.path.join(html_path, 'index.html'))

        js_file = pkgutil.get_data(__name__, "templates/js/app.js")
        write_byte(js_file, os.path.join(js_path, 'app.js'))

        css_file = pkgutil.get_data(__name__, "templates/css/app.css")
        write_byte(css_file, os.path.join(css_path, 'app.css'))

    def _map_static_files(self):
        self.app.static('/static', os.path.join(self.basepath,
                                                'assets'), name='static')
        self.app.url_for('static', filename='any')



# -*- coding: utf-8 -*-

__author__ = "Ngoc Huynh Bao"
__email__ = "ngoc.huynh.bao@nmbu.no"
__version__ = "0.0.1"


import pytest
from deoxys.keras.callbacks import Callback
from deoxys.model.callbacks import Callbacks
from deoxys.customize import register_callback, \
    unregister_callback, custom_callback
from deoxys.utils import Singleton


@pytest.fixture(autouse=True)
def clear_singleton():
    Singleton._instances = {}  # clear singleton


@pytest.fixture
def callback_class():
    class TestCallback(Callback):
        pass

    yield TestCallback


def test_is_singleton():
    callbacks_instance = Callbacks()
    another_instance = Callbacks()

    assert callbacks_instance is another_instance


def test_register_random_obj():
    with pytest.raises(ValueError):
        register_callback("TestCallback", object)


def test_register_callback_success(callback_class):
    register_callback("TestCallback", callback_class)

    assert Callbacks()._callbacks["TestCallback"] is callback_class


def test_register_duplicate_callback(callback_class):
    register_callback("TestCallback", callback_class)

    with pytest.raises(KeyError):
        register_callback("TestCallback", callback_class)


def test_unregister_callback(callback_class):
    register_callback("TestCallback", callback_class)

    assert Callbacks()._callbacks["TestCallback"] is callback_class
    unregister_callback("TestCallback")

    assert "TestCallback" not in Callbacks()._callbacks


def test_decorator():
    @custom_callback
    class TestCallback2(Callback):
        pass

    assert Callbacks()._callbacks["TestCallback2"] is TestCallback2
